import 'dart:ui';
import 'package:flutter/material.dart';
import 'package:flutter/rendering.dart';
import 'package:syncfusion_flutter_core/core.dart';
import '../../chart/chart_series/xy_data_series.dart';
import '../../chart/common/data_label.dart';
import '../../chart/utils/enum.dart';
import '../../chart/utils/helper.dart';
import '../../common/event_args.dart';
import '../../common/utils/helper.dart';

import '../../pyramid_chart/utils/helper.dart';
import '../base/circular_base.dart';
import '../base/circular_state_properties.dart';
import '../utils/enum.dart';
import '../utils/helper.dart';
import 'chart_point.dart';
import 'circular_chart_annotation.dart';
import 'renderer_extension.dart';

/// Represents the circular data label renderer
// ignore: must_be_immutable
class CircularDataLabelRenderer extends StatefulWidget {
  /// Creates the instance of circular data label renderer
  // ignore: prefer_const_constructors_in_immutables
  CircularDataLabelRenderer(
      {required this.stateProperties, required this.show});

  /// Specifies the circular chart state
  final CircularStateProperties stateProperties;

  /// Specifies whether to the data label
  bool show;

  /// Specifies the state of circular data label renderer
  late CircularDataLabelRendererState state;

  @override
  State<StatefulWidget> createState() {
    return CircularDataLabelRendererState();
  }
}

/// Represents the circular data label renderer state
class CircularDataLabelRendererState extends State<CircularDataLabelRenderer>
    with SingleTickerProviderStateMixin {
  /// Specifies the animation controller list
  late List<AnimationController> animationControllersList;

  /// Animation controller for series
  late AnimationController animationController;

  /// Repaint notifier for crosshair container
  late ValueNotifier<int> dataLabelRepaintNotifier;

  @override
  void initState() {
    dataLabelRepaintNotifier = ValueNotifier<int>(0);
    animationController = AnimationController(vsync: this)
      ..addListener(repaintDataLabelElements);
    super.initState();
  }

  @override
  Widget build(BuildContext context) {
    widget.state = this;
    animationController.duration = Duration(
        milliseconds:
            widget.stateProperties.renderingDetails.initialRender! ? 500 : 0);
    final Animation<double> dataLabelAnimation =
        Tween<double>(begin: 0.0, end: 1.0).animate(CurvedAnimation(
      parent: animationController,
      curve: const Interval(0.1, 0.8, curve: Curves.decelerate),
    ));
    animationController.forward(from: 0.0);
    return !widget.show
        ? Container()
        : Container(
            child: RepaintBoundary(
                child: CustomPaint(
                    painter: _CircularDataLabelPainter(
                        stateProperties: widget.stateProperties,
                        animation: dataLabelAnimation,
                        state: this,
                        notifier: dataLabelRepaintNotifier,
                        animationController: animationController))));
  }

  @override
  void dispose() {
    disposeAnimationController(animationController, repaintDataLabelElements);
    super.dispose();
  }

  /// Method to notify the repaint notifier
  void repaintDataLabelElements() {
    dataLabelRepaintNotifier.value++;
  }

  /// Method to render the data label
  void render() {
    setState(() {
      widget.show = true;
    });
  }
}

class _CircularDataLabelPainter extends CustomPainter {
  _CircularDataLabelPainter(
      {required this.stateProperties,
      required this.state,
      required this.animationController,
      required this.animation,
      required ValueNotifier<num> notifier})
      : super(repaint: notifier);

  final CircularStateProperties stateProperties;

  final CircularDataLabelRendererState state;

  final AnimationController animationController;

  final Animation<double> animation;

  /// To paint data labels
  @override
  void paint(Canvas canvas, Size size) {
    final List<CircularSeriesRendererExtension> visibleSeriesRenderers =
        stateProperties.chartSeries.visibleSeriesRenderers;
    CircularSeriesRendererExtension seriesRenderer;
    for (int seriesIndex = 0;
        seriesIndex < visibleSeriesRenderers.length;
        seriesIndex++) {
      seriesRenderer = visibleSeriesRenderers[seriesIndex];
      // ignore: unnecessary_null_comparison
      if (seriesRenderer.series.dataLabelSettings != null &&
          seriesRenderer.series.dataLabelSettings.isVisible) {
        seriesRenderer.dataLabelSettingsRenderer =
            DataLabelSettingsRenderer(seriesRenderer.series.dataLabelSettings);
        renderCircularDataLabel(
            seriesRenderer, canvas, stateProperties, seriesIndex, animation);
      }
    }
  }

  @override
  bool shouldRepaint(_CircularDataLabelPainter oldDelegate) => true;
}

/// Decides to increase the angle or not.
bool isIncreaseAngle = false;

/// To store the points which render at left and positioned outside
List<ChartPoint<dynamic>> leftPoints = <ChartPoint<dynamic>>[];

/// To store the points which render at right and positioned outside
List<ChartPoint<dynamic>> rightPoints = <ChartPoint<dynamic>>[];

/// Decrease the angle of the label if it intersects with labels.
void _decreaseAngle(
    ChartPoint<dynamic> currentPoint,
    ChartPoint<dynamic> previousPoint,
    CircularSeriesRendererExtension seriesRenderer,
    bool isRightSide) {
  int count = 1;
  if (isRightSide) {
    while (isOverlap(currentPoint.labelRect, previousPoint.labelRect) ||
        (seriesRenderer.series.pointRadiusMapper != null &&
            (!((previousPoint.labelRect.height + previousPoint.labelRect.top) <
                currentPoint.labelRect.top)))) {
      int newAngle = PointHelper.getNewAngle(previousPoint)!.toInt() - count;
      if (newAngle < 0) {
        newAngle = 360 + newAngle;
      }
      if (newAngle <= 270 && newAngle >= 90) {
        newAngle = 270;
        isIncreaseAngle = true;
        break;
      }
      _changeLabelAngle(previousPoint, newAngle, seriesRenderer);
      count++;
    }
  } else {
    if (PointHelper.getNewAngle(currentPoint)! > 270) {
      _changeLabelAngle(currentPoint, 270, seriesRenderer);
      PointHelper.setNewAngle(previousPoint, 270);
    }
    while (isOverlap(currentPoint.labelRect, previousPoint.labelRect) ||
        (seriesRenderer.series.pointRadiusMapper != null &&
            ((currentPoint.labelRect.top + currentPoint.labelRect.height) >
                previousPoint.labelRect.bottom))) {
      int newAngle = PointHelper.getNewAngle(previousPoint)!.toInt() - count;
      if (!(newAngle <= 270 && newAngle >= 90)) {
        newAngle = 270;
        isIncreaseAngle = true;
        break;
      }
      _changeLabelAngle(previousPoint, newAngle, seriesRenderer);
      if (isOverlap(currentPoint.labelRect, previousPoint.labelRect) &&
          leftPoints.indexOf(previousPoint) == null &&
          (newAngle - 1 < 90 && newAngle - 1 > 270)) {
        _changeLabelAngle(currentPoint,
            PointHelper.getNewAngle(currentPoint)! + 1, seriesRenderer);
        _arrangeLeftSidePoints(seriesRenderer);
        break;
      }
      count++;
    }
  }
}

/// Increase the angle of the label if it intersects labels.
void _increaseAngle(
    ChartPoint<dynamic> currentPoint,
    ChartPoint<dynamic> nextPoint,
    CircularSeriesRendererExtension seriesRenderer,
    bool isRightSide) {
  int count = 1;
  if (isRightSide) {
    while (isOverlap(currentPoint.labelRect, nextPoint.labelRect) ||
        (seriesRenderer.series.pointRadiusMapper != null &&
            (!((currentPoint.labelRect.top + currentPoint.labelRect.height) <
                nextPoint.labelRect.top)))) {
      int newAngle = PointHelper.getNewAngle(nextPoint)!.toInt() + count;
      if (newAngle < 270 && newAngle > 90) {
        newAngle = 90;
        isIncreaseAngle = true;
        break;
      }
      _changeLabelAngle(nextPoint, newAngle, seriesRenderer);
      if (isOverlap(currentPoint.labelRect, nextPoint.labelRect) &&
          (newAngle + 1 > 90 && newAngle + 1 < 270) &&
          rightPoints.indexOf(nextPoint) == rightPoints.length - 1) {
        _changeLabelAngle(currentPoint,
            PointHelper.getNewAngle(currentPoint)! - 1, seriesRenderer);
        _arrangeRightSidePoints(seriesRenderer);
        break;
      }
      count++;
    }
  } else {
    while (isOverlap(currentPoint.labelRect, nextPoint.labelRect) ||
        (seriesRenderer.series.pointRadiusMapper != null &&
            (currentPoint.labelRect.top <
                (nextPoint.labelRect.top + nextPoint.labelRect.height)))) {
      int newAngle = PointHelper.getNewAngle(nextPoint)!.toInt() + count;
      if (!(newAngle < 270 && newAngle > 90)) {
        newAngle = 270;
        isIncreaseAngle = false;
        break;
      }
      _changeLabelAngle(nextPoint, newAngle, seriesRenderer);
      count++;
    }
  }
}

/// Change the label angle based on the given new angle.
void _changeLabelAngle(ChartPoint<dynamic> currentPoint, num newAngle,
    CircularSeriesRendererExtension seriesRenderer) {
  const String defaultConnectorLineLength = '10%';
  final DataLabelSettings dataLabel =
      seriesRenderer.dataLabelSettingsRenderer.dataLabelSettings;
  final Size textSize = measureText(currentPoint.text!, dataLabel.textStyle);
  final Path angleChangedConnectorPath = Path();
  final num connectorLength = percentToValue(
      dataLabel.connectorLineSettings.length ?? defaultConnectorLineLength,
      currentPoint.outerRadius!)!;
  final Offset startPoint =
      degreeToPoint(newAngle, currentPoint.outerRadius!, currentPoint.center!);
  final Offset endPoint = degreeToPoint(newAngle,
      currentPoint.outerRadius! + connectorLength, currentPoint.center!);
  angleChangedConnectorPath.moveTo(startPoint.dx, startPoint.dy);
  if (dataLabel.connectorLineSettings.type == ConnectorType.line) {
    angleChangedConnectorPath.lineTo(endPoint.dx, endPoint.dy);
  }
  seriesRenderer
          .renderPoints![seriesRenderer.renderPoints!.indexOf(currentPoint)]
          .labelRect =
      getDataLabelRect(
          currentPoint.dataLabelPosition,
          seriesRenderer.dataLabelSettingsRenderer.dataLabelSettings
              .connectorLineSettings.type,
          dataLabel.margin,
          angleChangedConnectorPath,
          endPoint,
          textSize)!;
  PointHelper.setLabelUpdated(currentPoint, 1);
  PointHelper.setNewAngle(currentPoint, newAngle);
}

/// Left side points alignment calculation.
void _arrangeLeftSidePoints(CircularSeriesRendererExtension seriesRenderer) {
  ChartPoint<dynamic> previousPoint;
  ChartPoint<dynamic> currentPoint;
  bool angleChanged = false;
  bool startFresh = false;
  for (int i = 1; i < leftPoints.length; i++) {
    currentPoint = leftPoints[i];
    previousPoint = leftPoints[i - 1];
    if (isOverlapWithPrevious(currentPoint, leftPoints, i) &&
            currentPoint.isVisible ||
        !(PointHelper.getNewAngle(currentPoint)! < 270)) {
      angleChanged = true;
      if (startFresh) {
        isIncreaseAngle = false;
      }
      if (!isIncreaseAngle) {
        for (int k = i; k > 0; k--) {
          _decreaseAngle(
              leftPoints[k], leftPoints[k - 1], seriesRenderer, false);
          for (int index = 1; index < leftPoints.length; index++) {
            if (PointHelper.getLabelUpdated(leftPoints[index]) != null &&
                PointHelper.getNewAngle(leftPoints[index])! - 10 < 100) {
              isIncreaseAngle = true;
            }
          }
        }
      } else {
        for (int k = i; k < leftPoints.length; k++) {
          _increaseAngle(
              leftPoints[k - 1], leftPoints[k], seriesRenderer, false);
        }
      }
    } else {
      if (angleChanged &&
          previousPoint != null &&
          PointHelper.getLabelUpdated(previousPoint) == 1) {
        startFresh = true;
      }
    }
  }
}

/// Right side points alignments calculation.
void _arrangeRightSidePoints(CircularSeriesRendererExtension series) {
  bool startFresh = false;
  bool angleChanged = false;
  num checkAngle;
  ChartPoint<dynamic> currentPoint;
  final ChartPoint<dynamic>? lastPoint =
      rightPoints.length > 1 ? rightPoints[rightPoints.length - 1] : null;
  ChartPoint<dynamic> nextPoint;
  if (lastPoint != null) {
    if (PointHelper.getNewAngle(lastPoint)! > 360) {
      PointHelper.setNewAngle(
          lastPoint, PointHelper.getNewAngle(lastPoint)! - 360);
    }
    if (PointHelper.getNewAngle(lastPoint)! > 90 &&
        PointHelper.getNewAngle(lastPoint)! < 270) {
      isIncreaseAngle = true;
      _changeLabelAngle(lastPoint, 89, series);
    }
  }
  for (int i = rightPoints.length - 2; i >= 0; i--) {
    currentPoint = rightPoints[i];
    nextPoint = rightPoints[i + 1];
    if (isOverlapWithNext(currentPoint, rightPoints, i) &&
            currentPoint.isVisible ||
        !(PointHelper.getNewAngle(currentPoint)! <= 90 ||
            PointHelper.getNewAngle(currentPoint)! >= 270)) {
      checkAngle = PointHelper.getNewAngle(lastPoint!)! + 1;
      angleChanged = true;
      // If last's point change angle in beyond the limit, stop the increasing angle and do decrease the angle.
      if (startFresh) {
        isIncreaseAngle = false;
      } else if (checkAngle > 90 &&
          checkAngle < 270 &&
          PointHelper.getLabelUpdated(nextPoint) == 1) {
        isIncreaseAngle = true;
      }
      if (!isIncreaseAngle) {
        for (int k = i + 1; k < rightPoints.length; k++) {
          _increaseAngle(rightPoints[k - 1], rightPoints[k], series, true);
        }
      } else {
        for (int k = i + 1; k > 0; k--) {
          _decreaseAngle(rightPoints[k], rightPoints[k - 1], series, true);
        }
      }
    } else {
      //If a point did not overlapped with previous points, increase the angle always for right side points.
      if (angleChanged &&
          nextPoint != null &&
          PointHelper.getLabelUpdated(nextPoint) == 1) {
        startFresh = true;
      }
    }
  }
}

/// To render data label
void renderCircularDataLabel(
    CircularSeriesRendererExtension seriesRenderer,
    Canvas canvas,
    CircularStateProperties stateProperties,
    int seriesIndex,
    Animation<double>? animation) {
  ChartPoint<dynamic> point;
  final SfCircularChart chart = stateProperties.chart;
  final DataLabelSettings dataLabel = seriesRenderer.series.dataLabelSettings;
  final DataLabelSettingsRenderer dataLabelSettingsRenderer =
      seriesRenderer.dataLabelSettingsRenderer;
  final num angle = dataLabel.angle;
  Offset labelLocation;
  const int labelPadding = 2;
  String? label;
  final double animateOpacity = animation != null ? animation.value : 1;
  DataLabelRenderArgs dataLabelArgs;
  TextStyle dataLabelStyle;
  final List<Rect> renderDataLabelRegions = <Rect>[];
  Size textSize;
  for (int pointIndex = 0;
      pointIndex < seriesRenderer.renderPoints!.length;
      pointIndex++) {
    point = seriesRenderer.renderPoints![pointIndex];
    if (point.isVisible && (point.y != 0 || dataLabel.showZeroValue)) {
      label = point.text;
      label = seriesRenderer.renderer.getLabelContent(
          seriesRenderer, point, pointIndex, seriesIndex, label!);
      dataLabelStyle = dataLabel.textStyle;
      dataLabelSettingsRenderer.color =
          seriesRenderer.series.dataLabelSettings.color;
      if (chart.onDataLabelRender != null &&
          !seriesRenderer.renderPoints![pointIndex].labelRenderEvent) {
        dataLabelArgs = DataLabelRenderArgs(seriesRenderer,
            seriesRenderer.renderPoints, pointIndex, pointIndex);
        dataLabelArgs.text = label;
        dataLabelArgs.textStyle = dataLabelStyle;
        dataLabelArgs.color = dataLabelSettingsRenderer.color;
        chart.onDataLabelRender!(dataLabelArgs);
        label = point.text = dataLabelArgs.text;
        dataLabelStyle = dataLabelArgs.textStyle;
        pointIndex = dataLabelArgs.pointIndex!;
        dataLabelSettingsRenderer.color = dataLabelArgs.color;
        seriesRenderer.dataPoints[pointIndex].labelRenderEvent = true;
      }
      textSize = measureText(label, dataLabelStyle);

      /// condition check for labels after event.
      if (label != '') {
        if (seriesRenderer.seriesType == 'radialbar') {
          dataLabelStyle = chart.onDataLabelRender == null
              ? seriesRenderer.renderer.getDataLabelStyle(
                  seriesRenderer,
                  point,
                  pointIndex,
                  seriesIndex,
                  dataLabelStyle,
                  stateProperties.chartState)
              : dataLabelStyle;
          labelLocation = degreeToPoint(point.startAngle!,
              (point.innerRadius! + point.outerRadius!) / 2, point.center!);
          labelLocation = Offset(
              (labelLocation.dx - textSize.width - 5) +
                  (angle == 0 ? 0 : textSize.width / 2),
              (labelLocation.dy - textSize.height / 2) +
                  (angle == 0 ? 0 : textSize.height / 2));
          point.labelRect = Rect.fromLTWH(
              labelLocation.dx - labelPadding,
              labelLocation.dy - labelPadding,
              textSize.width + (2 * labelPadding),
              textSize.height + (2 * labelPadding));
          drawLabel(
              point.labelRect,
              labelLocation,
              label,
              null,
              canvas,
              seriesRenderer,
              point,
              pointIndex,
              seriesIndex,
              chart,
              dataLabelStyle,
              renderDataLabelRegions,
              animateOpacity);
        } else {
          setLabelPosition(
              dataLabel,
              point,
              textSize,
              stateProperties,
              canvas,
              renderDataLabelRegions,
              pointIndex,
              label,
              seriesRenderer,
              animateOpacity,
              dataLabelStyle,
              seriesIndex);
        }
      }
      dataLabelStyle = chart.onDataLabelRender == null
          ? seriesRenderer.renderer.getDataLabelStyle(
              seriesRenderer,
              point,
              pointIndex,
              seriesIndex,
              dataLabelStyle,
              stateProperties.chartState)
          : dataLabelStyle;
    } else {
      point.labelRect = Rect.zero;
    }
  }
  if (seriesRenderer.dataLabelSettingsRenderer.dataLabelSettings
              .labelIntersectAction ==
          LabelIntersectAction.shift &&
      seriesRenderer.seriesType != 'radialbar') {
    const int labelPadding = 2;
    leftPoints = <ChartPoint<dynamic>>[];
    rightPoints = <ChartPoint<dynamic>>[];
    for (int i = 0; i < seriesRenderer.renderPoints!.length; i++) {
      if (seriesRenderer.renderPoints![i].isVisible) {
        PointHelper.setNewAngle(seriesRenderer.renderPoints![i],
            seriesRenderer.renderPoints![i].midAngle);
        if (seriesRenderer.renderPoints![i].dataLabelPosition ==
                Position.left &&
            seriesRenderer.renderPoints![i].renderPosition ==
                ChartDataLabelPosition.outside) {
          leftPoints.add(seriesRenderer.renderPoints![i]);
        } else if (seriesRenderer.renderPoints![i].dataLabelPosition ==
                Position.right &&
            seriesRenderer.renderPoints![i].renderPosition ==
                ChartDataLabelPosition.outside) {
          rightPoints.add(seriesRenderer.renderPoints![i]);
        }
      }
    }
    leftPoints.sort((ChartPoint<dynamic> a, ChartPoint<dynamic> b) =>
        PointHelper.getNewAngle(a)!.compareTo(PointHelper.getNewAngle(b)!));
    if (leftPoints.isNotEmpty) {
      _arrangeLeftSidePoints(seriesRenderer);
    }
    isIncreaseAngle = false;
    if (rightPoints.isNotEmpty) {
      _arrangeRightSidePoints(seriesRenderer);
    }
    for (int pointIndex = 0;
        pointIndex < seriesRenderer.renderPoints!.length;
        pointIndex++) {
      if (seriesRenderer.renderPoints![pointIndex].isVisible) {
        final ChartPoint<dynamic> point =
            seriesRenderer.renderPoints![pointIndex];
        final EdgeInsets margin =
            seriesRenderer.series.dataLabelSettings.margin;
        final Rect rect = point.labelRect;
        TextStyle dataLabelStyle = dataLabel.textStyle;
        dataLabelStyle = TextStyle(
            color: (chart.onDataLabelRender != null && dataLabelSettingsRenderer.color != null)
                ? getSaturationColor(
                    dataLabelSettingsRenderer.color ?? point.fill)
                : ((dataLabelStyle.color ?? dataLabel.textStyle.color) ??
                    getSaturationColor(
                        point.renderPosition == ChartDataLabelPosition.outside
                            ? findthemecolor(stateProperties, point, dataLabel)
                            : dataLabelSettingsRenderer.color ?? point.fill)),
            fontSize: dataLabelStyle.fontSize ?? dataLabel.textStyle.fontSize,
            fontFamily:
                dataLabelStyle.fontFamily ?? dataLabel.textStyle.fontFamily,
            fontStyle:
                dataLabelStyle.fontStyle ?? dataLabel.textStyle.fontStyle,
            fontWeight:
                dataLabelStyle.fontWeight ?? dataLabel.textStyle.fontWeight,
            inherit: dataLabelStyle.inherit,
            backgroundColor: dataLabelStyle.backgroundColor ??
                dataLabel.textStyle.backgroundColor,
            letterSpacing: dataLabelStyle.letterSpacing ??
                dataLabel.textStyle.letterSpacing,
            wordSpacing:
                dataLabelStyle.wordSpacing ?? dataLabel.textStyle.wordSpacing,
            textBaseline:
                dataLabelStyle.textBaseline ?? dataLabel.textStyle.textBaseline,
            height: dataLabelStyle.height ?? dataLabel.textStyle.height,
            locale: dataLabelStyle.locale ?? dataLabel.textStyle.locale,
            foreground:
                dataLabelStyle.foreground ?? dataLabel.textStyle.foreground,
            background:
                dataLabelStyle.background ?? dataLabel.textStyle.background,
            shadows: dataLabelStyle.shadows ?? dataLabel.textStyle.shadows,
            fontFeatures:
                dataLabelStyle.fontFeatures ?? dataLabel.textStyle.fontFeatures,
            decoration:
                dataLabelStyle.decoration ?? dataLabel.textStyle.decoration,
            decorationColor: dataLabelStyle.decorationColor ??
                dataLabel.textStyle.decorationColor,
            decorationStyle: dataLabelStyle.decorationStyle ?? dataLabel.textStyle.decorationStyle,
            decorationThickness: dataLabelStyle.decorationThickness ?? dataLabel.textStyle.decorationThickness,
            debugLabel: dataLabelStyle.debugLabel ?? dataLabel.textStyle.debugLabel,
            fontFamilyFallback: dataLabelStyle.fontFamilyFallback ?? dataLabel.textStyle.fontFamilyFallback);
        textSize = measureText(label!, dataLabelStyle);
        labelLocation = Offset(
            rect.left +
                (point.renderPosition == ChartDataLabelPosition.inside
                    ? labelPadding
                    : margin.left),
            rect.top + rect.height / 2 - textSize.height / 2);
        const String defaultConnectorLineLength = '10%';
        point.trimmedText = point.text;
        Path shiftedConnectorPath = Path();
        final num connectorLength = percentToValue(
            seriesRenderer.dataLabelSettingsRenderer.dataLabelSettings
                    .connectorLineSettings.length ??
                defaultConnectorLineLength,
            point.outerRadius!)!;
        final Offset startPoint = degreeToPoint(
            (point.startAngle! + point.endAngle!) / 2,
            point.outerRadius!,
            point.center!);
        final Offset endPoint = degreeToPoint(PointHelper.getNewAngle(point)!,
            point.outerRadius! + connectorLength, point.center!);
        shiftedConnectorPath.moveTo(startPoint.dx, startPoint.dy);
        if (seriesRenderer.dataLabelSettingsRenderer.dataLabelSettings
                .connectorLineSettings.type ==
            ConnectorType.line) {
          shiftedConnectorPath.lineTo(endPoint.dx, endPoint.dy);
        }
        getDataLabelRect(
            point.dataLabelPosition,
            seriesRenderer.dataLabelSettingsRenderer.dataLabelSettings
                .connectorLineSettings.type,
            margin,
            shiftedConnectorPath,
            endPoint,
            textSize);
        final ChartLocation midAngle = getPerpendicularDistance(
            ChartLocation(startPoint.dx, startPoint.dy), point);
        if (seriesRenderer.dataLabelSettingsRenderer.dataLabelSettings
                    .connectorLineSettings.type ==
                ConnectorType.curve &&
            PointHelper.getLabelUpdated(point) == 1) {
          const int spacing = 10;
          shiftedConnectorPath = Path();
          shiftedConnectorPath.moveTo(startPoint.dx, startPoint.dy);
          shiftedConnectorPath.quadraticBezierTo(
              midAngle.x,
              midAngle.y,
              endPoint.dx -
                  (point.dataLabelPosition == Position.left
                      ? spacing
                      : -spacing),
              endPoint.dy);
        }
        final Rect containerRect =
            stateProperties.renderingDetails.chartAreaRect;
        if (containerRect.left > rect.left) {
          labelLocation = Offset(containerRect.left,
              rect.top + rect.height / 2 - textSize.height / 2);
        }
        if (point.labelRect.left < containerRect.left &&
            point.renderPosition == ChartDataLabelPosition.outside) {
          point.trimmedText = getTrimmedText(point.trimmedText!,
              point.labelRect.right - containerRect.left, dataLabelStyle, null);
        }
        if (point.labelRect.right > containerRect.right &&
            point.renderPosition == ChartDataLabelPosition.outside) {
          point.trimmedText = getTrimmedText(point.trimmedText!,
              containerRect.right - point.labelRect.left, dataLabelStyle, null);
        }
        if (point.trimmedText != '' &&
            !isOverlapWithPrevious(
                point, seriesRenderer.renderPoints!, pointIndex) &&
            rect != Rect.zero) {
          drawLabel(
              rect,
              labelLocation,
              point.trimmedText!,
              point.renderPosition == ChartDataLabelPosition.outside
                  ? shiftedConnectorPath
                  : Path(),
              canvas,
              seriesRenderer,
              point,
              pointIndex,
              seriesIndex,
              chart,
              dataLabelStyle,
              renderDataLabelRegions,
              animateOpacity);
        }
      }
    }
  }
}

/// To set data label position
void setLabelPosition(
    DataLabelSettings dataLabel,
    ChartPoint<dynamic> point,
    Size textSize,
    CircularStateProperties stateProperties,
    Canvas canvas,
    List<Rect> renderDataLabelRegions,
    int pointIndex,
    String label,
    CircularSeriesRendererExtension seriesRenderer,
    double animateOpacity,
    TextStyle dataLabelStyle,
    int seriesIndex) {
  final DataLabelSettingsRenderer dataLabelSettingsRenderer =
      seriesRenderer.dataLabelSettingsRenderer;
  final SfCircularChart chart = stateProperties.chart;
  final num angle = dataLabel.angle;
  Offset labelLocation;
  final bool smartLabel = seriesRenderer.series.enableSmartLabels;
  const int labelPadding = 2;
  if (dataLabel.labelPosition == ChartDataLabelPosition.inside) {
    labelLocation = degreeToPoint(point.midAngle!,
        (point.innerRadius! + point.outerRadius!) / 2, point.center!);
    labelLocation = Offset(
        labelLocation.dx -
            (textSize.width / 2) +
            (angle == 0 ? 0 : textSize.width / 2),
        labelLocation.dy -
            (textSize.height / 2) +
            (angle == 0 ? 0 : textSize.height / 2));
    point.labelRect = Rect.fromLTWH(
        labelLocation.dx - labelPadding,
        labelLocation.dy - labelPadding,
        textSize.width + (2 * labelPadding),
        textSize.height + (2 * labelPadding));
    final bool isDataLabelCollide =
        findingCollision(point.labelRect, renderDataLabelRegions);
    final TextStyle textStyle = TextStyle(
        color: (dataLabelStyle.color ?? dataLabel.textStyle.color) ??
            getSaturationColor(
                findthemecolor(stateProperties, point, dataLabel)),
        fontSize: dataLabelStyle.fontSize ?? dataLabel.textStyle.fontSize,
        fontFamily: dataLabelStyle.fontFamily ?? dataLabel.textStyle.fontFamily,
        fontStyle: dataLabelStyle.fontStyle ?? dataLabel.textStyle.fontStyle,
        fontWeight: dataLabelStyle.fontWeight ?? dataLabel.textStyle.fontWeight,
        inherit: dataLabelStyle.inherit,
        backgroundColor: dataLabelStyle.backgroundColor ??
            dataLabel.textStyle.backgroundColor,
        letterSpacing:
            dataLabelStyle.letterSpacing ?? dataLabel.textStyle.letterSpacing,
        wordSpacing:
            dataLabelStyle.wordSpacing ?? dataLabel.textStyle.wordSpacing,
        textBaseline:
            dataLabelStyle.textBaseline ?? dataLabel.textStyle.textBaseline,
        height: dataLabelStyle.height ?? dataLabel.textStyle.height,
        locale: dataLabelStyle.locale ?? dataLabel.textStyle.locale,
        foreground: dataLabelStyle.foreground ?? dataLabel.textStyle.foreground,
        background: dataLabelStyle.background ?? dataLabel.textStyle.background,
        shadows: dataLabelStyle.shadows ?? dataLabel.textStyle.shadows,
        fontFeatures:
            dataLabelStyle.fontFeatures ?? dataLabel.textStyle.fontFeatures,
        decoration: dataLabelStyle.decoration ?? dataLabel.textStyle.decoration,
        decorationColor: dataLabelStyle.decorationColor ??
            dataLabel.textStyle.decorationColor,
        decorationStyle: dataLabelStyle.decorationStyle ??
            dataLabel.textStyle.decorationStyle,
        decorationThickness: dataLabelStyle.decorationThickness ??
            dataLabel.textStyle.decorationThickness,
        debugLabel: dataLabelStyle.debugLabel ?? dataLabel.textStyle.debugLabel,
        fontFamilyFallback: dataLabelStyle.fontFamilyFallback ??
            dataLabel.textStyle.fontFamilyFallback);
    if (seriesRenderer.series.dataLabelSettings.labelIntersectAction ==
            LabelIntersectAction.shift &&
        isDataLabelCollide) {
      point.saturationRegionOutside = true;
      point.renderPosition = ChartDataLabelPosition.outside;
      dataLabelStyle = textStyle;
      renderOutsideDataLabel(
          canvas,
          label,
          point,
          textSize,
          pointIndex,
          seriesRenderer,
          seriesIndex,
          stateProperties,
          dataLabelStyle,
          renderDataLabelRegions,
          animateOpacity);
    } else if (smartLabel && isDataLabelCollide) {
      point.saturationRegionOutside = true;
      point.renderPosition = ChartDataLabelPosition.outside;
      dataLabelStyle = textStyle;
      renderOutsideDataLabel(
          canvas,
          label,
          point,
          textSize,
          pointIndex,
          seriesRenderer,
          seriesIndex,
          stateProperties,
          dataLabelStyle,
          renderDataLabelRegions,
          animateOpacity);
    } else {
      point.renderPosition = ChartDataLabelPosition.inside;
      dataLabelStyle = TextStyle(
          color: (chart.onDataLabelRender != null &&
                  dataLabelSettingsRenderer.color != null)
              ? getSaturationColor(
                  dataLabelSettingsRenderer.color ?? point.fill)
              : ((dataLabelStyle.color ?? dataLabel.textStyle.color) ??
                  getSaturationColor(
                      dataLabelSettingsRenderer.color ?? point.fill)),
          fontSize: dataLabelStyle.fontSize ?? dataLabel.textStyle.fontSize,
          fontFamily:
              dataLabelStyle.fontFamily ?? dataLabel.textStyle.fontFamily,
          fontStyle: dataLabelStyle.fontStyle ?? dataLabel.textStyle.fontStyle,
          fontWeight:
              dataLabelStyle.fontWeight ?? dataLabel.textStyle.fontWeight,
          inherit: dataLabelStyle.inherit,
          backgroundColor: dataLabelStyle.backgroundColor ??
              dataLabel.textStyle.backgroundColor,
          letterSpacing:
              dataLabelStyle.letterSpacing ?? dataLabel.textStyle.letterSpacing,
          wordSpacing:
              dataLabelStyle.wordSpacing ?? dataLabel.textStyle.wordSpacing,
          textBaseline:
              dataLabelStyle.textBaseline ?? dataLabel.textStyle.textBaseline,
          height: dataLabelStyle.height ?? dataLabel.textStyle.height,
          locale: dataLabelStyle.locale ?? dataLabel.textStyle.locale,
          foreground:
              dataLabelStyle.foreground ?? dataLabel.textStyle.foreground,
          background:
              dataLabelStyle.background ?? dataLabel.textStyle.background,
          shadows: dataLabelStyle.shadows ?? dataLabel.textStyle.shadows,
          fontFeatures:
              dataLabelStyle.fontFeatures ?? dataLabel.textStyle.fontFeatures,
          decoration:
              dataLabelStyle.decoration ?? dataLabel.textStyle.decoration,
          decorationColor: dataLabelStyle.decorationColor ??
              dataLabel.textStyle.decorationColor,
          decorationStyle: dataLabelStyle.decorationStyle ??
              dataLabel.textStyle.decorationStyle,
          decorationThickness: dataLabelStyle.decorationThickness ??
              dataLabel.textStyle.decorationThickness,
          debugLabel:
              dataLabelStyle.debugLabel ?? dataLabel.textStyle.debugLabel,
          fontFamilyFallback: dataLabelStyle.fontFamilyFallback ??
              dataLabel.textStyle.fontFamilyFallback);
      if (!isDataLabelCollide &&
          (dataLabel.labelIntersectAction == LabelIntersectAction.shift)) {
        renderDataLabelRegions.add(point.labelRect);
      } else if (!isDataLabelCollide &&
          (dataLabel.labelIntersectAction == LabelIntersectAction.hide)) {
        drawLabel(
            point.labelRect,
            labelLocation,
            label,
            null,
            canvas,
            seriesRenderer,
            point,
            pointIndex,
            seriesIndex,
            chart,
            dataLabelStyle,
            renderDataLabelRegions,
            animateOpacity);
      } else {
        drawLabel(
            point.labelRect,
            labelLocation,
            label,
            null,
            canvas,
            seriesRenderer,
            point,
            pointIndex,
            seriesIndex,
            chart,
            dataLabelStyle,
            renderDataLabelRegions,
            animateOpacity);
      }
    }
  } else {
    point.renderPosition = ChartDataLabelPosition.outside;
    dataLabelStyle = TextStyle(
        color: (dataLabelStyle.color ?? dataLabel.textStyle.color) ??
            getSaturationColor(
                findthemecolor(stateProperties, point, dataLabel)),
        fontSize: dataLabelStyle.fontSize ?? dataLabel.textStyle.fontSize,
        fontFamily: dataLabelStyle.fontFamily ?? dataLabel.textStyle.fontFamily,
        fontStyle: dataLabelStyle.fontStyle ?? dataLabel.textStyle.fontStyle,
        fontWeight: dataLabelStyle.fontWeight ?? dataLabel.textStyle.fontWeight,
        inherit: dataLabelStyle.inherit,
        backgroundColor: dataLabelStyle.backgroundColor ??
            dataLabel.textStyle.backgroundColor,
        letterSpacing:
            dataLabelStyle.letterSpacing ?? dataLabel.textStyle.letterSpacing,
        wordSpacing:
            dataLabelStyle.wordSpacing ?? dataLabel.textStyle.wordSpacing,
        textBaseline:
            dataLabelStyle.textBaseline ?? dataLabel.textStyle.textBaseline,
        height: dataLabelStyle.height ?? dataLabel.textStyle.height,
        locale: dataLabelStyle.locale ?? dataLabel.textStyle.locale,
        foreground: dataLabelStyle.foreground ?? dataLabel.textStyle.foreground,
        background: dataLabelStyle.background ?? dataLabel.textStyle.background,
        shadows: dataLabelStyle.shadows ?? dataLabel.textStyle.shadows,
        fontFeatures:
            dataLabelStyle.fontFeatures ?? dataLabel.textStyle.fontFeatures,
        decoration: dataLabelStyle.decoration ?? dataLabel.textStyle.decoration,
        decorationColor: dataLabelStyle.decorationColor ??
            dataLabel.textStyle.decorationColor,
        decorationStyle: dataLabelStyle.decorationStyle ??
            dataLabel.textStyle.decorationStyle,
        decorationThickness: dataLabelStyle.decorationThickness ??
            dataLabel.textStyle.decorationThickness,
        debugLabel: dataLabelStyle.debugLabel ?? dataLabel.textStyle.debugLabel,
        fontFamilyFallback: dataLabelStyle.fontFamilyFallback ??
            dataLabel.textStyle.fontFamilyFallback);
    renderOutsideDataLabel(
        canvas,
        label,
        point,
        textSize,
        pointIndex,
        seriesRenderer,
        seriesIndex,
        stateProperties,
        dataLabelStyle,
        renderDataLabelRegions,
        animateOpacity);
  }
}

/// To render outside positioned data labels.
void renderOutsideDataLabel(
    Canvas canvas,
    String label,
    ChartPoint<dynamic> point,
    Size textSize,
    int pointIndex,
    CircularSeriesRendererExtension seriesRenderer,
    int seriesIndex,
    CircularStateProperties stateProperties,
    TextStyle textStyle,
    List<Rect> renderDataLabelRegions,
    double animateOpacity) {
  Path connectorPath;
  Rect? rect;
  Offset labelLocation;
  const String defaultConnectorLineLength = '10%';
  final EdgeInsets margin = seriesRenderer.series.dataLabelSettings.margin;
  final ConnectorLineSettings connector =
      seriesRenderer.series.dataLabelSettings.connectorLineSettings;
  connectorPath = Path();
  final num connectorLength = percentToValue(
      connector.length ?? defaultConnectorLineLength, point.outerRadius!)!;
  final Offset startPoint =
      degreeToPoint(point.midAngle!, point.outerRadius!, point.center!);
  final Offset endPoint = degreeToPoint(
      point.midAngle!, point.outerRadius! + connectorLength, point.center!);
  connectorPath.moveTo(startPoint.dx, startPoint.dy);
  if (connector.type == ConnectorType.line) {
    connectorPath.lineTo(endPoint.dx, endPoint.dy);
  }
  rect = getDataLabelRect(point.dataLabelPosition, connector.type, margin,
      connectorPath, endPoint, textSize);
  point.labelRect = rect!;
  labelLocation = Offset(rect.left + margin.left,
      rect.top + rect.height / 2 - textSize.height / 2);
  final Rect containerRect = stateProperties.renderingDetails.chartAreaRect;
  if (seriesRenderer.series.dataLabelSettings.builder == null) {
    if (seriesRenderer.series.dataLabelSettings.labelIntersectAction ==
        LabelIntersectAction.hide) {
      if (!findingCollision(rect, renderDataLabelRegions) &&
          (rect.left > containerRect.left &&
              rect.left + rect.width <
                  containerRect.left + containerRect.width) &&
          rect.top > containerRect.top &&
          rect.top + rect.height < containerRect.top + containerRect.height) {
        drawLabel(
            rect,
            labelLocation,
            label,
            connectorPath,
            canvas,
            seriesRenderer,
            point,
            pointIndex,
            seriesIndex,
            stateProperties.chart,
            textStyle,
            renderDataLabelRegions,
            animateOpacity);
      }
    } else if (seriesRenderer.series.dataLabelSettings.labelIntersectAction ==
        LabelIntersectAction.shift) {
      renderDataLabelRegions.add(rect);
    } else {
      drawLabel(
          rect,
          labelLocation,
          label,
          connectorPath,
          canvas,
          seriesRenderer,
          point,
          pointIndex,
          seriesIndex,
          stateProperties.chart,
          textStyle,
          renderDataLabelRegions,
          animateOpacity);
    }
  } else {
    if (seriesRenderer.series.dataLabelSettings.labelIntersectAction !=
        LabelIntersectAction.shift)
      canvas.drawPath(
          connectorPath,
          Paint()
            ..color = connector.width <= 0
                ? Colors.transparent
                : connector.color ?? point.fill.withOpacity(animateOpacity)
            ..strokeWidth = connector.width
            ..style = PaintingStyle.stroke);
  }
}

/// To draw label
void drawLabel(
    Rect labelRect,
    Offset location,
    String label,
    Path? connectorPath,
    Canvas canvas,
    CircularSeriesRendererExtension seriesRenderer,
    ChartPoint<dynamic> point,
    int pointIndex,
    int seriesIndex,
    SfCircularChart chart,
    TextStyle textStyle,
    List<Rect> renderDataLabelRegions,
    double animateOpacity) {
  Paint rectPaint;
  final DataLabelSettings dataLabel = seriesRenderer.series.dataLabelSettings;
  final DataLabelSettingsRenderer dataLabelSettingsRenderer =
      seriesRenderer.dataLabelSettingsRenderer;
  final ConnectorLineSettings connector = dataLabel.connectorLineSettings;
  if (connectorPath != null) {
    canvas.drawPath(
        connectorPath,
        Paint()
          ..color = connector.width <= 0
              ? Colors.transparent
              : connector.color ?? point.fill.withOpacity(animateOpacity)
          ..strokeWidth = connector.width
          ..style = PaintingStyle.stroke);
  }

  if (dataLabel.builder == null) {
    final double strokeWidth = seriesRenderer.renderer.getDataLabelStrokeWidth(
        seriesRenderer, point, pointIndex, seriesIndex, dataLabel.borderWidth);
    final Color? labelFill = seriesRenderer.renderer.getDataLabelColor(
        seriesRenderer,
        point,
        pointIndex,
        seriesIndex,
        dataLabelSettingsRenderer.color ??
            (dataLabel.useSeriesColor
                ? point.fill
                : dataLabelSettingsRenderer.color));
    final Color strokeColor = seriesRenderer.renderer.getDataLabelStrokeColor(
        seriesRenderer,
        point,
        pointIndex,
        seriesIndex,
        dataLabel.borderColor.withOpacity(dataLabel.opacity));
    // ignore: unnecessary_null_comparison
    if (strokeWidth != null && strokeWidth > 0) {
      rectPaint = Paint()
        ..color = strokeColor.withOpacity(
            (animateOpacity - (1 - dataLabel.opacity)) < 0
                ? 0
                : animateOpacity - (1 - dataLabel.opacity))
        ..style = PaintingStyle.stroke
        ..strokeWidth = strokeWidth;
      drawLabelRect(
          rectPaint,
          Rect.fromLTRB(
              labelRect.left, labelRect.top, labelRect.right, labelRect.bottom),
          dataLabel.borderRadius,
          canvas);
    }
    if (labelFill != null) {
      drawLabelRect(
          Paint()
            ..color = labelFill.withOpacity(
                (animateOpacity - (1 - dataLabel.opacity)) < 0
                    ? 0
                    : animateOpacity - (1 - dataLabel.opacity))
            ..style = PaintingStyle.fill,
          Rect.fromLTRB(
              labelRect.left, labelRect.top, labelRect.right, labelRect.bottom),
          dataLabel.borderRadius,
          canvas);
    }
    drawText(canvas, label, location, textStyle, dataLabel.angle);
    if (seriesRenderer.series.dataLabelSettings.labelIntersectAction !=
        LabelIntersectAction.shift) {
      renderDataLabelRegions.add(labelRect);
    }
  }
}

/// Method to trigger the data label event
void triggerCircularDataLabelEvent(
    SfCircularChart chart,
    CircularSeriesRendererExtension seriesRenderer,
    CircularStateProperties stateProperties,
    Offset? position) {
  const int seriesIndex = 0;
  final DataLabelSettings dataLabel = seriesRenderer.series.dataLabelSettings;
  Offset labelLocation;
  num connectorLength;
  ChartPoint<dynamic> point;
  const String defaultConnectorLineLength = '10%';
  for (int index = 0; index < seriesRenderer.dataPoints.length; index++) {
    point = seriesRenderer.dataPoints[index];
    if (dataLabel.isVisible &&
        // ignore: unnecessary_null_comparison
        seriesRenderer.dataPoints[index].labelRect != null &&
        position != null &&
        seriesRenderer.dataPoints[index].labelRect.contains(position)) {
      if (dataLabel.labelPosition == ChartDataLabelPosition.inside) {
        labelLocation = degreeToPoint(point.midAngle!,
            (point.innerRadius! + point.outerRadius!) / 2, point.center!);
        position = Offset(labelLocation.dx, labelLocation.dy);
      } else {
        connectorLength = percentToValue(
            dataLabel.connectorLineSettings.length ??
                defaultConnectorLineLength,
            point.outerRadius!)!;
        labelLocation = degreeToPoint(point.midAngle!,
            point.outerRadius! + connectorLength, point.center!);
        position = Offset(labelLocation.dx, labelLocation.dy);
      }
      if (chart.onDataLabelTapped != null) {
        dataLabelTapEvent(chart, seriesRenderer.series.dataLabelSettings, index,
            point, position, seriesIndex);
      }
    }
  }
}

/// To draw data label rect
void drawLabelRect(
        Paint paint, Rect labelRect, double borderRadius, Canvas canvas) =>
    canvas.drawRRect(
        RRect.fromRectAndRadius(labelRect, Radius.circular(borderRadius)),
        paint);

/// To find data label position
void findDataLabelPosition(ChartPoint<dynamic> point) {
  point.midAngle =
      point.midAngle! > 360 ? point.midAngle! - 360 : point.midAngle!;
  point.dataLabelPosition = ((point.midAngle! >= -90 && point.midAngle! < 0) ||
          (point.midAngle! >= 0 && point.midAngle! < 90) ||
          (point.midAngle! >= 270))
      ? Position.right
      : Position.left;
}

/// Method for setting color to data label
Color findthemecolor(CircularStateProperties stateProperties,
    ChartPoint<dynamic> point, DataLabelSettings dataLabel) {
  return dataLabel.color ??
      (dataLabel.useSeriesColor
          ? point.fill
          : (stateProperties.chart.backgroundColor ??
              (stateProperties.renderingDetails.chartTheme.brightness ==
                      Brightness.light
                  ? Colors.white
                  : Colors.black)));
}
